@@ -15,23 +15,23 @@
 #  author: terry.yinzhe@gmail.com
 #
 """
-lizard is a simple code complexity analyzer without caring about the C/C++ 
+lizard is a simple code complexity analyzer without caring about the C/C++
 header files or Java imports. It can deal with
 
 * Java
 * C/C++
 * Objective C.
 
-It counts 
+It counts
 
-* the nloc (net lines of code, excluding comments and blanks), 
+* the nloc (net lines of code, excluding comments and blanks),
 * CCN (cyclomatic complexity number),
 * token count of functions.
 * parameter count of functions.
 
 You can set limitation for CCN (-C), the number of parameters (-a). Functions
 that exceed these limitations will generate warnings. The exit code of lizard
-will be none-Zero if there are warnings. 
+will be none-Zero if there are warnings.
 
 This tool actually calculates how complex the code 'looks' rather than how
 complex the code real 'is'. People will need this tool because it's often very
@@ -42,12 +42,13 @@ complexity.
 It requires python2.6 or above (early versions are not verified).
 """
 
+BUG_REPORTING = "please report bug to terry.yinzhe@gmail.com or https://github.com/terryyin/lizard.\n"
 
-VERSION = "0.0.1"
-HFCCA_VERSION = "1.7.2"
+VERSION = "0.0.2"
+HFCCA_VERSION = "1.7.3"
 
 
-import itertools
+import itertools, traceback
 import re
 
 
@@ -57,25 +58,26 @@ DEFAULT_CCN_THRESHOLD = 15
 '''
 The analyzing process
 =====================
-Source code => Tokens => UniversalCode
+Source code => Tokens => SourceCodeInforamtionBuilder
 
-UniversalCode.get_statistics => The statistics of one source file.
+SourceCodeInforamtionBuilder.get_statistics => The statistics of one source file.
 
 '''
 
 
 class FunctionInfo(object):
-    ''' 
+    '''
     Statistic information of a function.
     '''
-    
-    def __init__(self, name, start_line):
-        self.cyclomatic_complexity = 1
+
+    def __init__(self, name, start_line = 0, ccn = 1):
+        self.cyclomatic_complexity = ccn
         self.nloc = 0
-        self.token_count = 0
+        self.token_count = 1  # the first token
         self.name = name
         self.long_name = name
         self.start_line = start_line
+        self.end_line = None
         self.parameter_count = 0
 
     def add_to_function_name(self, app):
@@ -87,7 +89,7 @@ class FunctionInfo(object):
 
     def add_parameter(self, token):
         self.add_to_long_name(" " + token)
-        
+
         if self.parameter_count == 0:
             self.parameter_count = 1
         if token == ",":
@@ -99,11 +101,12 @@ class FileInformation(object):
     Statistic information of a source file.
     Including all the functions and the file summary.
     '''
-    
+
     def __init__(self, filename, nloc, function_list):
         self.filename = filename
         self.nloc = nloc
         self.function_list = function_list
+        self.token_count = 0
 
     average_NLOC = property(lambda self:self._functions_average("nloc"))
     average_token = property(lambda self:self._functions_average("token_count"))
@@ -115,10 +118,10 @@ class FileInformation(object):
                 / len(self.function_list) if self.function_list else 0
 
 
-class UniversalCode(object):
+class SourceCodeInforamtionBuilder(object):
     """
-        UniversalCode is the code that is unrelated to any programming
-        languages. The code could be:
+        SourceCodeInforamtionBuilder provides the builder to build FileInformation
+        The building blocks are:
         START_NEW_FUNCTION
             ADD_TO_FUNCTION_NAME
             ADD_TO_LONG_FUNCTION_NAME
@@ -126,21 +129,20 @@ class UniversalCode(object):
                     CONDITION
                     TOKEN
         END_OF_FUNCTION
-
-        A TokenTranslator will generate UniversalCode.
     """
 
     def __init__(self):
-        self.current_function = None
-        self.newline = True
-        self.nloc = 0
-        self.function_list = []
+        self.fileinfo = FileInformation("", 0, [])
+        self.START_NEW_FUNCTION('', 0)
+
+    def build(self):
+        return self.fileinfo
 
     def START_NEW_FUNCTION(self, name, start_line):
+        self.newline = True
         self.current_function = FunctionInfo(name, start_line)
 
     def CONDITION(self):
-        self.TOKEN()
         self.current_function.cyclomatic_complexity += 1
 
     def TOKEN(self):
@@ -150,7 +152,7 @@ class UniversalCode(object):
         self.current_function.token_count += 1
 
     def NEW_LINE(self):
-        self.nloc += 1
+        self.fileinfo.nloc += 1
         self.newline = True
 
     def ADD_TO_LONG_FUNCTION_NAME(self, app):
@@ -162,8 +164,27 @@ class UniversalCode(object):
     def PARAMETER(self, token):
         self.current_function.add_parameter(token)
 
-    def END_OF_FUNCTION(self):
-        self.function_list.append(self.current_function)
+    def END_OF_FUNCTION(self, end_line):
+        self.current_function.end_line = end_line
+        self.fileinfo.function_list.append(self.current_function)
+        self.START_NEW_FUNCTION('', 0)
+
+
+class ParsingError(Exception):
+
+    def __init__(self, line_number):
+        self.line_number = line_number
+        self.filename = ''
+        self.code = ''
+
+    def __str__(self):
+        try:
+            line = self.code.splitlines()[self.line_number - 1]
+        except:
+            line = ''
+        return '''!!!Exception Happens!!!
+At %s:%d: '%s'
+If possible, ''' % (self.filename, self.line_number, line) + BUG_REPORTING
 
 
 class LanguageReaderBase(object):
@@ -171,7 +192,7 @@ class LanguageReaderBase(object):
     This is the base class for any programming language reader.
     It reader the tokens of a certain language and generate universal_code
     by calling it's -generate_universal_code- method.
-    
+
     The derived class should implement a series of 'state' functions. The
     most basic and start state is _GLOBAL. Derived class must implement _GLOBAL
     and other states. When the state changes, simply assign the state method to
@@ -183,15 +204,27 @@ class LanguageReaderBase(object):
     def __init__(self):
         self._state = self._GLOBAL
         self.current_line = 0
-        self.univeral_code = UniversalCode()
-    
+        self.sourceCodeInfoBuilder = SourceCodeInforamtionBuilder()
+
     def generate_universal_code(self, tokens):
         for token, self.current_line in tokens:
-            if token.isspace():
-                self.univeral_code.NEW_LINE()
-            else:
-                self._state(token)
-        return self.univeral_code
+            try:
+                if token.isspace():
+                    self.sourceCodeInfoBuilder.NEW_LINE()
+                else:
+                    self.sourceCodeInfoBuilder.TOKEN()
+                    self.sourceCodeInfoBuilder.fileinfo.token_count+=1
+                    self.process_token(token)
+            except:
+                raise ParsingError(self.current_line)
+
+        return self.sourceCodeInfoBuilder
+
+    def process_token(self, token):
+        return self._state(token)
+
+    def extend_tokens(self, tokens):
+        return tokens
 
 
 class CLikeReader(LanguageReaderBase):
@@ -202,37 +235,60 @@ class CLikeReader(LanguageReaderBase):
     def __init__(self):
         super(CLikeReader, self).__init__()
         self.conditions = set(
-            ['if', 'for', 'while', '&&', '||', 'case', '?', '#if', 'catch'])
+            ['if', 'for', 'while', '&&', '||', 'case', '?', '#if', '#elif', 'catch'])
         self.bracket_level = 0
         self.br_count = 0
         self.last_preprocessor = None
 
+    def extend_tokens(self, tokens):
+        last_token_is_include = False
+        include_brackets = False
+        brackets = ""
+        for token, line in tokens:
+            if last_token_is_include and token == "<":
+                include_brackets = True
+                brackets = "<"
+            elif include_brackets:
+                brackets += token
+                if token == ">":
+                    include_brackets = False
+                    token = brackets
+            last_token_is_include = token == "#include"
+            if not include_brackets:
+                yield token, line
+
     def remove_hash_if_from_conditions(self):
         self.conditions.remove("#if")
+        self.conditions.remove("#elif")
 
     def _is_condition(self, token):
         return token in self.conditions
 
+    def process_token(self, token):
+        if token.startswith("#") and self._state != self._IMP:
+            return
+        return super(CLikeReader, self).process_token(token)
+
     def _GLOBAL(self, token):
         if token == '(':
             self.bracket_level += 1
             self._state = self._DEC
-            self.univeral_code.ADD_TO_LONG_FUNCTION_NAME(token)
+            self.sourceCodeInfoBuilder.ADD_TO_LONG_FUNCTION_NAME(token)
         elif token == '::':
             self._state = self._NAMESPACE
         else:
+            self.sourceCodeInfoBuilder.START_NEW_FUNCTION(token, self.current_line)
             if token == 'operator':
                 self._state = self._OPERATOR
-            self.univeral_code.START_NEW_FUNCTION(token, self.current_line)
 
     def _OPERATOR(self, token):
         if token != '(':
             self._state = self._GLOBAL
-        self.univeral_code.ADD_TO_FUNCTION_NAME(' ' + token)
+        self.sourceCodeInfoBuilder.ADD_TO_FUNCTION_NAME(' ' + token)
 
     def _NAMESPACE(self, token):
         self._state = self._OPERATOR if token == 'operator'  else self._GLOBAL
-        self.univeral_code.ADD_TO_FUNCTION_NAME("::" + token)
+        self.sourceCodeInfoBuilder.ADD_TO_FUNCTION_NAME("::" + token)
 
     def _DEC(self, token):
         if token in ('(', "<"):
@@ -242,19 +298,26 @@ class CLikeReader(LanguageReaderBase):
             if (self.bracket_level == 0):
                 self._state = self._DEC_TO_IMP
         elif self.bracket_level == 1:
-            self.univeral_code.PARAMETER(token)
+            self.sourceCodeInfoBuilder.PARAMETER(token)
             return
-        self.univeral_code.ADD_TO_LONG_FUNCTION_NAME(" " + token)
+        self.sourceCodeInfoBuilder.ADD_TO_LONG_FUNCTION_NAME(" " + token)
 
     def _DEC_TO_IMP(self, token):
         if token == 'const':
-            self.univeral_code.ADD_TO_LONG_FUNCTION_NAME(" " + token)
+            self.sourceCodeInfoBuilder.ADD_TO_LONG_FUNCTION_NAME(" " + token)
         elif token == '{':
             self.br_count += 1
             self._state = self._IMP
+        elif token == ":":
+            self._state = self._CONSTRUCTOR_INITIALIZATION_LIST
         else:
             self._state = self._GLOBAL
 
+    def _CONSTRUCTOR_INITIALIZATION_LIST(self, token):
+        if token == '{':
+            self.br_count += 1
+            self._state = self._IMP
+
     def _IMP(self, token):
         if token in ("#else", "#if", "#endif"):
             self.last_preprocessor = token
@@ -266,12 +329,10 @@ class CLikeReader(LanguageReaderBase):
                 self.br_count -= 1
                 if self.br_count == 0:
                     self._state = self._GLOBAL
-                    self.univeral_code.END_OF_FUNCTION()
+                    self.sourceCodeInfoBuilder.END_OF_FUNCTION(self.current_line)
                     return
         if self._is_condition(token):
-            self.univeral_code.CONDITION()
-        else:
-            self.univeral_code.TOKEN()
+            self.sourceCodeInfoBuilder.CONDITION()
 
 
 class ObjCReader(CLikeReader):
@@ -285,12 +346,12 @@ class ObjCReader(CLikeReader):
             super(ObjCReader, self)._DEC_TO_IMP(token)
             if self._state == self._GLOBAL:
                 self._state = self._OBJC_DEC_BEGIN
-                self.univeral_code.START_NEW_FUNCTION(token, self.current_line)
+                self.sourceCodeInfoBuilder.START_NEW_FUNCTION(token, self.current_line)
 
     def _OBJC_DEC_BEGIN(self, token):
         if token == ':':
             self._state = self._OBJC_DEC
-            self.univeral_code.ADD_TO_FUNCTION_NAME(token)
+            self.sourceCodeInfoBuilder.ADD_TO_FUNCTION_NAME(token)
         elif token == '{':
             self.br_count += 1
             self._state = self._IMP
@@ -300,7 +361,7 @@ class ObjCReader(CLikeReader):
     def _OBJC_DEC(self, token):
         if token == '(':
             self._state = self._OBJC_PARAM_TYPE
-            self.univeral_code.ADD_TO_LONG_FUNCTION_NAME(token)
+            self.sourceCodeInfoBuilder.ADD_TO_LONG_FUNCTION_NAME(token)
         elif token == ',':
             pass
         elif token == '{':
@@ -308,41 +369,43 @@ class ObjCReader(CLikeReader):
             self._state = self._IMP
         else:
             self._state = self._OBJC_DEC_BEGIN
-            self.univeral_code.ADD_TO_FUNCTION_NAME(" " + token)
+            self.sourceCodeInfoBuilder.ADD_TO_FUNCTION_NAME(" " + token)
 
     def _OBJC_PARAM_TYPE(self, token):
         if token == ')':
             self._state = self._OBJC_PARAM
-        self.univeral_code.ADD_TO_LONG_FUNCTION_NAME(" " + token)
+        self.sourceCodeInfoBuilder.ADD_TO_LONG_FUNCTION_NAME(" " + token)
 
     def _OBJC_PARAM(self, token):
         self._state = self._OBJC_DEC
 
 
+def compile_file_extension_re(*exts):
+    return re.compile(r".*\.(" + r"|".join(exts) + r")$", re.IGNORECASE)
+
 class LanguageChooser(object):
 
     lizard_language_infos = {
                      'c/c++': {
-                          'name_pattern': re.compile(r".*\.(c|C|cpp|CPP|CC|cc|mm)$"),
+                          'name_pattern': compile_file_extension_re("c", "cpp", "cc", "mm", "cxx", "h", "hpp"),
                           'reader':CLikeReader},
                      'Java': {
-                          'name_pattern': re.compile(r".*\.(java)$"),
+                          'name_pattern': compile_file_extension_re("java"),
                           'reader':CLikeReader},
                       'objC' : {
-                          'name_pattern': re.compile(r".*\.(m)$"),
+                          'name_pattern': compile_file_extension_re("m"),
                           'reader':ObjCReader}
                     }
 
-    def is_file_type_supported(self, filename):
-        return any(info['name_pattern'].match(filename) 
-                   for info in self.lizard_language_infos.values())
-
-    def get_reader_by_file_name_otherwise_default(self, filename):
+    def get_language_by_filename(self, filename):
         for lan in self.lizard_language_infos:
             info = self.lizard_language_infos[lan]
             if info['name_pattern'].match(filename):
-                return info['reader']()
-        return self.lizard_language_infos['c/c++']['reader']()
+                return lan
+
+    def get_reader_by_file_name_otherwise_default(self, filename):
+        lan = self.get_language_by_filename(filename)
+        return self.lizard_language_infos[lan or "c/c++"]['reader']()
 
 
 class FileAnalyzer:
@@ -350,57 +413,125 @@ class FileAnalyzer:
         Returns a list of function infos in this file.
     '''
 
-    def __init__(self, noCountPre=False):
+    def __init__(self, extensions = [], noCountPre=False):
         self.no_preprocessor_count = noCountPre
+        self.extensions = extensions
 
     def __call__(self, filename):
-        with open(filename) as f:
-            return self.analyze_source_code(filename, f.read())
+        try:
+            with open(filename) as f:
+                return self.analyze_source_code(filename, f.read())
+        except Exception as e:
+            msg = '\n'.join(traceback.format_exception_only(type(e), e))
+            msg+= "\nIf you think this is a bug, "+ BUG_REPORTING
+            sys.stderr.write(msg)
 
     def analyze_source_code(self, filename, code):
-        reader = LanguageChooser().get_reader_by_file_name_otherwise_default(filename)
-        if self.no_preprocessor_count and hasattr(reader, 'remove_hash_if_from_conditions'):
-            reader.remove_hash_if_from_conditions()
-        result = self.analyze_source_code_with_parser(filename, code, reader)
-        return result
+        try:
+            reader = LanguageChooser().get_reader_by_file_name_otherwise_default(filename)
+            if self.no_preprocessor_count and hasattr(reader, 'remove_hash_if_from_conditions'):
+                reader.remove_hash_if_from_conditions()
+            result = self.analyze_source_code_with_parser(filename, code, reader)
+
+            return result
+        except ParsingError as e:
+            e.filename = filename
+            e.code = code
+            raise
 
     def analyze_source_code_with_parser(self, filename, code, parser):
         tokens = generate_tokens(code)
-        parsed_code = parser.generate_universal_code(tokens)
-        function_list = parsed_code.function_list
-        return FileInformation(filename, parsed_code.nloc, function_list)
-
+        tokens = parser.extend_tokens(tokens)
+        for extension in self.extensions:
+            tokens = extension.extend_tokens(tokens)
+        result = parser.generate_universal_code(tokens).build()
+        result.filename = filename
+        for extension in self.extensions:
+            result.wordCount = extension.result
+        return result
 
 analyze_file = FileAnalyzer()
 
 
-token_pattern = re.compile(r"(\w+|/\*|//|:=|::|>=|\*=|\*\*|\*|>"+
+def warning_filter(option, fileStatistics):
+    for file_info in fileStatistics:
+        if file_info:
+            for fun in file_info.function_list:
+                if fun.cyclomatic_complexity > option.CCN or \
+                        fun.parameter_count > option.arguments:
+                    yield fun, file_info.filename
+
+
+class Whitelist(object):
+
+    def __init__(self, text):
+        self.whitelist = [
+            self._get_whitelist_item(line.split('#')[0])
+            for line in text.splitlines()]
+
+    def filter(self, warnings):
+        for warning in warnings:
+            if not self._in_list(warning):
+                yield warning
+
+    def _get_whitelist_item(self, text):
+        white = {}
+        pieces = text.replace('::', '##').split(':')
+        if len(pieces) > 1:
+            white['file_name'] = pieces[0]
+            text = pieces[1]
+        white['function_names'] = [x.strip().replace('##','::') for x in text.split(',')]
+        return white
+
+    def _in_list(self, warning):
+        return any(self._match_whitelist_item(white, warning) for white in self.whitelist)
+
+    def _match_whitelist_item(self, white, warning):
+        if warning[0].name in white['function_names']:
+            if 'file_name' in white:
+                return warning[1] == white['file_name']
+            return True
+        return False
+
+class FreeFormattingTokenizer(object):
+    '''
+    Use this tokenizer to tokenize C/C++, Java, ObjC code, which the
+    format is not part of the syntax. So indentation & new lines
+    doesn't matter.
+    '''
+
+    token_pattern = re.compile(r"(\w+|/\*|//|:=|::|>=|\*=|\*\*|\*|>"+
                            r"|&=|&&|&"+
                            r"|#\s*define|#\s*if|#\s*else|#\s*endif|#\s*\w+"+
                            r"|[!%^&\*\-=+\|\\<>/\]\+]+|.)", re.M | re.S)
 
-def generate_tokens(source_code):
-    for t, l in generate_tokens_from_code(source_code):
-        if not any(t.startswith(x) for x in ('#define', '/*', '//')) :
-            yield t, l
 
-def generate_tokens_from_code(source_code):
-    in_middle_of_empty_lines = False
-    for (token, line) in tokens_from_code_with_multiple_newlines(source_code):
-        if token != '\n' or not in_middle_of_empty_lines:
-            yield token, line
-        in_middle_of_empty_lines = (token == '\n')
+    def __call__(self, source_code):
+        for t, l in self._generate_tokens_without_empty_lines(source_code):
+            if not any(t.startswith(x) for x in ('#define', '/*', '//')) :
+                yield t, l
 
-def tokens_from_code_with_multiple_newlines(source_code):
-    index = 0
-    line = 1
-    while 1:
-        m = token_pattern.match(source_code, index)
-        if not m:
-            break
-        token = m.group(0)
-        if token == '\n': line += 1
+    def _generate_tokens_without_empty_lines(self, source_code):
+        in_middle_of_empty_lines = False
+        for (token, line) in self._tokens_from_code_with_multiple_newlines(source_code):
+            if token != '\n' or not in_middle_of_empty_lines:
+                yield token, line
+            in_middle_of_empty_lines = (token == '\n')
 
+    def _tokens_from_code_with_multiple_newlines(self, source_code):
+        index = 0
+        line = 1
+        while index >= 0:
+            m = self.token_pattern.match(source_code, index)
+            if not m:
+                break
+            index, token = self._read_one_token(source_code, index, m.group(0))            
+            line += 1 if token == '\n' else (len(token.splitlines()) - 1)
+            if not token.isspace() or token == '\n':
+                yield token, line
+
+    def _read_one_token(self, source_code, index, token):
+        original_index = index
         if token.startswith("#"):
             token = "#" + token[1:].strip()
 
@@ -412,19 +543,15 @@ def tokens_from_code_with_multiple_newlines(source_code):
                     break
                 if not source_code[bindex:index].rstrip().endswith('\\'):
                     break
-            if index == -1:
-                break
-            token = source_code[m.start(0):index]
+            if index != -1:
+                token = source_code[original_index:index]
         elif token == '/*':
             index = source_code.find("*/", index + 2)
-            if index == -1:
-                break
-            index += 2
-            token = source_code[m.start(0):index]
+            if index != -1:
+                index += 2
+                token = source_code[original_index:index]
         elif token in( '//', '#if','#endif'):
             index = source_code.find('\n', index)
-            if index == -1:
-                break
         elif token == '"' or token == '\'':
             while(1):
                 index += 1
@@ -435,15 +562,15 @@ def tokens_from_code_with_multiple_newlines(source_code):
                     if source_code[index - 2] != '\\':
                         continue
                 break
-            if index == -1:
-                break
-            token = source_code[m.start(0):index + 1]
-            index = index + 1
+            if index != -1:
+                token = source_code[original_index:index + 1]
+                index = index + 1
         else:
-            index = m.end(0)
-        line += (len(token.splitlines()) - 1)
-        if not token.isspace() or token == '\n':
-            yield token, line
+            index += len(token)
+        return index, token
+
+generate_tokens = FreeFormattingTokenizer()
+
 
 import sys
 
@@ -453,13 +580,14 @@ def print_function_info_header():
     print("--------------------------------------------------------------")
 
 def print_function_info(fun, filename, option):
+    line = '{}-{}'.format(fun.start_line, fun.end_line) if option.display_fn_end_line else fun.start_line
     output_params = {
         'nloc': fun.nloc,
         'CCN': fun.cyclomatic_complexity,
         'token': fun.token_count,
         'param': fun.parameter_count,
         'name': fun.name,
-        'line': fun.start_line,
+        'line': line,
         'file': filename
     }
     output_format = "%(nloc)6d %(CCN)6d %(token)6d %(param)6d    %(name)s@%(line)s@%(file)s"
@@ -469,19 +597,16 @@ def print_function_info(fun, filename, option):
         output_format = "%(file)s:%(line)s: warning: %(name)s has %(CCN)d CCN and %(param)d params (%(nloc)d NLOC, %(token)d tokens)"
     print(output_format % output_params)
 
-def print_warnings(option, saved_result):
+def print_warnings(option, warnings):
     warning_count = 0
     if not option.warnings_only:
         print(("\n" +
                "======================================\n" +
-              "!!!! Warnings (CCN > %d) !!!!\n" +
-              "======================================") % option.CCN)
-    for file_info in saved_result:
-        for fun in file_info.function_list:
-            if fun.cyclomatic_complexity > option.CCN or \
-                    fun.parameter_count > option.arguments:
-                warning_count += 1
-                print_function_info(fun, file_info.filename, option)
+              "!!!! Warnings (CCN > %d) !!!!") % option.CCN)
+        print_function_info_header()
+    for warning in warnings:
+        warning_count += 1
+        print_function_info(warning[0], warning[1], option)
 
     if warning_count == 0:
         print("No warning found. Excellent!")
@@ -489,7 +614,7 @@ def print_warnings(option, saved_result):
     return warning_count
 
 def print_total(warning_count, saved_result, option):
-    file_infos = list(saved_result)
+    file_infos = list(file_info for file_info in saved_result if file_info)
     all_fun = list(itertools.chain(*(file_info.function_list for file_info in file_infos)))
     cnt = len(all_fun)
     if (cnt == 0):
@@ -522,9 +647,12 @@ def print_and_save_detail_information(allStatistics, option):
     else:
         print_function_info_header()
         for fileStatistics in allStatistics:
-            all_functions.append(fileStatistics)
-            for fun in fileStatistics.function_list:
-                print_function_info(fun, fileStatistics.filename, option)
+            for extension in option.extensions:
+                extension.reduce(fileStatistics)
+            if fileStatistics:
+                all_functions.append(fileStatistics)
+                for fun in fileStatistics.function_list:
+                    print_function_info(fun, fileStatistics.filename, option)
 
         print("--------------------------------------------------------------")
         print("%d file analyzed." % (len(all_functions)))
@@ -544,32 +672,39 @@ def print_and_save_detail_information(allStatistics, option):
 
 def print_result(r, option):
     all_functions = print_and_save_detail_information(r, option)
-    warning_count = print_warnings(option, all_functions)
+    warnings = warning_filter(option, all_functions)
+    warnings = Whitelist(option.whitelist).filter(warnings)
+    warning_count = print_warnings(option, warnings)
     print_total(warning_count, all_functions, option)
-    if option.number > warning_count:
+    for extension in option.extensions:
+        extension.print_result()
+    if option.number < warning_count:
         sys.exit(1)
 
 class XMLFormatter(object):
-    
+
     def xml_output(self, result, options):
         ''' Thanks for Holy Wen from Nokia Siemens Networks to let me use his code
             to put the result into xml file that is compatible with cppncss.
             Jenkens has plugin for cppncss format result to display the diagram.
         '''
         import xml.dom.minidom
-            
+
         impl = xml.dom.minidom.getDOMImplementation()
         doc = impl.createDocument(None, "cppncss", None)
         root = doc.documentElement
-    
+
+        pi = doc.createProcessingInstruction('xml-stylesheet','type="text/xsl" href="https://raw.github.com/terryyin/lizard/master/lizard.xsl"')
+        doc.insertBefore(pi, root)
+
         measure = doc.createElement("measure")
         measure.setAttribute("type", "Function")
         measure.appendChild(self._createLabels(doc, ["Nr.", "NCSS", "CCN"]))
-    
+
         Nr = 0
         total_func_ncss = 0
         total_func_ccn = 0
-    
+
         for source_file in result:
             file_name = source_file.filename
             for func in source_file.function_list:
@@ -577,44 +712,44 @@ class XMLFormatter(object):
                 total_func_ncss += func.nloc
                 total_func_ccn += func.cyclomatic_complexity
                 measure.appendChild(self._createFunctionItem(doc, Nr, file_name, func, options.verbose))
-    
+
             if Nr != 0:
                 measure.appendChild(self._createLabeledValueItem(doc, 'average', "NCSS", str(total_func_ncss / Nr)))
                 measure.appendChild(self._createLabeledValueItem(doc, 'average', "CCN", str(total_func_ccn / Nr)))
-    
+
         root.appendChild(measure)
-    
+
         measure = doc.createElement("measure")
         measure.setAttribute("type", "File")
         measure.appendChild(self._createLabels(doc, ["Nr.", "NCSS", "CCN", "Functions"]))
-    
+
         file_NR = 0
         file_total_ncss = 0
         file_total_ccn = 0
         file_total_funcs = 0
-    
+
         for source_file in result:
             file_NR += 1
             file_total_ncss += source_file.nloc
             file_total_ccn += source_file.CCN
             file_total_funcs += len(source_file.function_list)
             measure.appendChild(self._createFileNode(doc, source_file, file_NR))
-    
+
         if file_NR != 0:
             fileSummary = [("NCSS", file_total_ncss / file_NR),
                            ("CCN", file_total_ccn / file_NR),
                            ("Functions", file_total_funcs / file_NR)]
             for k, v in fileSummary:
                 measure.appendChild(self._createLabeledValueItem(doc, 'average', k, v))
-    
+
         summary = [("NCSS", file_total_ncss),
                        ("CCN", file_total_ccn ),
                        ("Functions", file_total_funcs)]
         for k, v in summary:
             measure.appendChild(self._createLabeledValueItem(doc, 'sum', k, v))
-    
+
         root.appendChild(measure)
-    
+
         xmlString = doc.toprettyxml()
         return xmlString
 
@@ -628,7 +763,7 @@ class XMLFormatter(object):
         labels = doc.createElement("labels")
         for label in labelNames:
             labels.appendChild(self._createLabel(doc, label))
-        
+
         return labels
 
     def _createFunctionItem(self, doc, Nr, file_name, func, verbose):
@@ -680,7 +815,7 @@ class XMLFormatter(object):
         item.appendChild(value4)
         return item
 
-def createlizardCommandLineParser():
+def createCommandLineParser():
     from optparse import OptionParser
     parser = OptionParser(version=VERSION)
     parser.add_option("-v", "--verbose",
@@ -733,6 +868,21 @@ def createlizardCommandLineParser():
             type="int",
             dest="working_threads",
             default=1)
+    parser.add_option("-d", "--find_duplicates",
+            help="find and skip analysis for identical files. Will be made default in the next release",
+            action="store_true",
+            dest="duplicates",
+            default=False)
+    parser.add_option("-e", "--display_fn_end_line",
+            help="display function end line number in addition to start line number. Will be made default in the next release",
+            action="store_true",
+            dest="display_fn_end_line",
+            default=False)
+    parser.add_option("-E", "--extension",
+            help="under construction...", #"Use extension. Can be WordCount.",
+            action="append",
+            dest="extensions",
+            default=[])
 
     parser.usage = "lizard [options] [PATH or FILE] [PATH] ... "
     parser.description = __doc__
@@ -756,34 +906,79 @@ def mapFilesToAnalyzer(files, fileAnalyzer, working_threads):
 
 import os
 import fnmatch
+import hashlib
 
-def _notExluded(str_to_match, patterns):
-    return LanguageChooser().is_file_type_supported(str_to_match) and \
-        all(not fnmatch.fnmatch(str_to_match, p) for p in patterns)
+class FilesFilter(object):
+    
+    def __init__(self, exclude_patterns, check_duplicates):
+        self.exclude_patterns = exclude_patterns
+        self.check_duplicates = check_duplicates
+        self.hash_set = set()
+        
+    def getFileNames(self, paths):
+        for SRC_DIR in paths:
+            if os.path.isfile(SRC_DIR) :
+                yield SRC_DIR
+            else:
+                for root, _, files in os.walk(SRC_DIR, topdown=False):
+                    for filename in files:
+                        full_path_name = os.path.join(root, filename)
+                        if self._checkFile(full_path_name):
+                            yield full_path_name
 
-def getSourceFiles(SRC_DIRs, exclude_patterns):
-    for SRC_DIR in SRC_DIRs:
-        if os.path.isfile(SRC_DIR) and LanguageChooser().is_file_type_supported(SRC_DIR):
-            yield SRC_DIR
-        else:
-            for root, _, files in os.walk(SRC_DIR, topdown=False):
-                for filename in files:
-                    full_path_name = os.path.join(root, filename)
-                    if _notExluded(full_path_name, exclude_patterns):
-                        yield full_path_name
+    def _checkFile(self, full_path_name):
+        if self._notExluded(full_path_name):
+            if self.check_duplicates:
+                return self._notDuplicate(full_path_name)
+            else:
+                return True
+
+    def _md5HashFile(self, full_path_name):
+        ''' return md5 hash of a file '''
+        with open(full_path_name, mode='r') as source_file:
+            if sys.version_info[0] == 3:
+                code_md5 = hashlib.md5(source_file.read().encode('utf-8'))
+            else:
+                code_md5 = hashlib.md5(source_file.read())
+        return code_md5.hexdigest()
+
+    def _notExluded(self, str_to_match):
+        return LanguageChooser().get_language_by_filename(str_to_match) and \
+            all(not fnmatch.fnmatch(str_to_match, p) for p in self.exclude_patterns)
+
+    def _notDuplicate(self, full_path_name):
+        ''' Function counts md5 hash for the given file 
+            and checks if it isn't a duplicate using set 
+            of hashes for previous files '''
+        fhash = self._md5HashFile(full_path_name)
+        if fhash and fhash not in self.hash_set:
+            self.hash_set.add(fhash)
+            return True
+
+def get_extensions(extension_names):
+    return [__import__('lizard' + name).LizardExtension() for name in extension_names]
 
 def analyze(paths, options):
     ''' This is the most important function of lizard.
         It analyze the given paths with the options.
         Can be used directly by other Python application.
     '''
-    files = getSourceFiles(paths, options.exclude)
-    fileAnalyzer = FileAnalyzer(options.no_preprocessor_count)
+    files = FilesFilter(options.exclude, options.duplicates).getFileNames(paths)
+    fileAnalyzer = FileAnalyzer(options.extensions, options.no_preprocessor_count)
     r = mapFilesToAnalyzer(files, fileAnalyzer, options.working_threads)
     return r
 
+def get_whitelist():
+    try:
+        with open('whitelizard.txt', mode='r') as whitelizard:
+            return whitelizard.read()
+    except:
+        return ''
+
 def lizard_main(argv):
-    options, args = createlizardCommandLineParser().parse_args(args=argv)
+    options, args = createCommandLineParser().parse_args(args=argv)
+    options.whitelist = get_whitelist()
+    options.extensions = get_extensions(options.extensions)
     paths = ["."] if len(args) == 1 else args[1:]
     r = analyze(paths, options)
     if options.xml:
