@@ -1,5 +1,5 @@
 import unittest
-from mock import patch
+from test.mock import patch
 import lizard
 from lizard import lizard_main
 import os
@@ -74,8 +74,23 @@ class TestOptions(unittest.TestCase):
     def test_not_with_preprocessor_counted_in_CCN(self):
         self.runApplicationWithArgv(['lizard', '-P'])
         self.assertEqual(1, self.fileInfos[0].function_list[0].cyclomatic_complexity)
+        
+    def test_using_the_WordCount_plugin(self):
+        self.runApplicationWithArgv(['lizard', '-EWordCount'])
+        self.assertEqual(1, self.fileInfos[0].wordCount["foo"])
 
 
-if __name__ == "__main__":
-    #import sys;sys.argv = ['', 'Test.testName']
-    unittest.main()
\ No newline at end of file
+from lizard import get_whitelist
+class TestOptionParsing(unittest.TestCase):
+
+    @patch('lizard.open', create=True)
+    def test_load_whitelist_from_file(self, mock_open):
+        mock_open.return_value.__enter__.return_value.read.return_value = "foo"
+        self.assertEqual("foo", get_whitelist())
+        mock_open.assert_called_with('whitelizard.txt', mode='r')
+
+    @patch('lizard.open', create=True)
+    def test_should_be_empty_if_whitelist_file_doesnot_exist(self, mock_open):
+        mock_open.side_effect = IOError
+        self.assertEqual("", get_whitelist())
+
